package BankServer;

import java.net.*;
import java.io.*;
import BankApp.*;

//import java.io.ByteArrayInputStream;
//import java.io.DataInputStream;

public class UDPServerSocket extends UDPSocket {
	public Packet[] receivedBuffer = new Packet[BankUtility.UDPReceiveBufferSize];
	
	public String socketLock = "";

	public int beginPos = 0;

	public int endPos = 0;

	public UDPServerListener listener = null;

	public UDPServerSocket(int portNumber) throws Exception {
		myPortNumber = portNumber;
		recvSocket = new DatagramSocket(myPortNumber);
		sendSocket = new DatagramSocket();// no port is needed
		recvSocket.setSoTimeout(BankUtility.WaitingLength);

		// Start the listener which receive packets and feed the buffer
		listener = new UDPServerListener(this);
		listener.start();
	}

	/*
	 * Clean up the socket
	 */
	public void close() {
		// Debug
		if(recvSocket!=null){
			if(!recvSocket.isClosed()){
				BankUtility.showDebugInfo(1,"UDPServerListener stopped on port: " + recvSocket.getLocalPort());
			}
		}

		// Close the sockets
		super.close();
		// Shutdown the listener
		if (listener != null) {
			listener.isListening = false;
			try {
				listener.join();
			}
			catch (Exception e) {
			}
		}
	}

	/**
	 * Retrieve packet from received buffer
	 * 
	 * @return The packet if one exists, or null if no packets received
	 * @throws Exception
	 */
	public Packet recvPacket() throws Exception {
		Packet result = null;
		// Check if there is packets in the buffer
		synchronized (this.receivedBuffer) {
			if (beginPos != endPos) {
				// There are packets
				result = receivedBuffer[beginPos++];
				if (beginPos > BankUtility.UDPReceiveBufferSize)
					beginPos = 0;
				// Return the packet
				return result;
			}
		}
		Thread.yield();

		synchronized (this.socketLock) {
			synchronized (this.receivedBuffer) {
				if (beginPos != endPos) {
					// There are packets, get one
					result = receivedBuffer[beginPos++];
					if (beginPos > BankUtility.UDPReceiveBufferSize)
						beginPos = 0;
				}
				else {
					// Still no packet, throw exception
					throw new ConnectException();
				}
			}
		}
		return result;
	}

	// (if we are going to monitor incoming message, there is no timeout,)
	// if we are acknowledge receipt, we must have this timeout, otherwise
	// system maybe deadlock
	public Packet recvPacketInner() throws Exception {
		boolean isTestPacket = true;
		Packet result = null;

		// Don't return the received packet if the packet is a test packet
		while (isTestPacket) {
			result = null;
			int index = 0;
			do {
				// recvSocket=new DatagramSocket(myPortNumber);
				recvSocket.setSoTimeout(BankUtility.WaitingLength);
				result = doRecvPacket();

				String theHostName = recvPacket.getAddress().getHostAddress();
				int thePortNumber = recvPacket.getPort();

				// debug
				BankUtility.showDebugInfo(1, "Received one packet from " + theHostName + ":" + thePortNumber + " Seq=" + result.seqNumber);

				Packet ack = new Packet();

				ack.opID = result.opID;

				ack.seqNumber = result.seqNumber;
				// ///////////////////////////////
				// BankUtility.ACKNOWLEDGEMENT;
				replyPortNumber = result.tag;

				// replyHostName = recvPacket.getAddress().getHostName();
				replyHostName = recvPacket.getAddress().getHostAddress();

				ack.status = BankUtility.ACKNOWLEDGEMENT;
				// make sure bankimplement has this

				if (result.status != BankUtility.ACKNOWLEDGEMENT) {
					// debug
					BankUtility.showDebugInfo(1, "Reply to " + recvPacket.getAddress().getHostAddress() + ":" + result.tag + " Seq="+ack.seqNumber);

					doSendPacket(recvPacket.getAddress().getHostAddress(), result.tag, ack);
				}

				index = getIndexByName(theHostName, thePortNumber);

				if (index == -1) {
					// this is a new connection
					BankUtility.showDebugInfo(1, "New connection:" + connectionCount + " on port "
							+ recvSocket.getLocalPort());
					index = connectionCount;
					connectionCount++;
					hostNames[index] = theHostName;
					portNumbers[index] = thePortNumber;
					seqNumber[index] = result.seqNumber;
				}
				// System.out.println("server call gethost by adddress and index="+ index);

			}// while (result.seqNumber < seqNumber[index]);
			while (result.seqNumber < seqNumber[index]);

			// seqNumber[index] = result.seqNumber;
			seqNumber[index] = result.seqNumber + 1;

			// System.out.println("server acknowledge OK");

			// Check if the packet is a test packet, if it's not, return the packet
			if (result.status != Packet.TEST)
				isTestPacket = false;
		} // End of while

		return result;
	}

} // end class

/**
 * The UDP Server Listener which cache the UDP packet received
 * 
 * @author Min Huang (khtwo)
 */
class UDPServerListener extends Thread {
	public boolean isListening = true;

	public UDPServerSocket theSocket = null;

	public Packet currentPacket = null;

	public UDPServerListener(UDPServerSocket socket) {
		theSocket = socket;
	}

	/*
	 * Receive packets and put them in the buffer
	 */
	public void run() {
		BankUtility.showDebugInfo(1,"UDPServerListener start on port: " + theSocket.recvSocket.getLocalPort());
		while (isListening) {
			try {
				synchronized (theSocket.socketLock) {
					// Receive a packet and put it into the buffer
					currentPacket = theSocket.recvPacketInner();
				}
				Thread.yield();

				// Check if the buffer is full
				synchronized (theSocket.receivedBuffer) {
					if (theSocket.endPos == theSocket.beginPos - 1
							|| (theSocket.beginPos == 0 && theSocket.endPos == BankUtility.UDPReceiveBufferSize)) {
						// Buffer full, ignor the newest packet
						continue;
					}
					theSocket.receivedBuffer[theSocket.endPos++] = currentPacket;
					// Rewind the buffer when overflow
					if (theSocket.endPos > BankUtility.UDPReceiveBufferSize) {
						theSocket.endPos = 0;
					}
				}
			}
			catch (Exception e) {
				// Time out, do nothing
				Thread.yield();
			}
		}
	}
}
